//
// webserver.c
//
// Simple HTTP server sample for sanos
//

#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <dirent.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define SERVER "webserver/1.0"
#define PROTOCOL "HTTP/1.0"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"
#define PORT 80

//============================================================================

void __fprintf(int fd, const char *fmt, ...)
{
    static char buf[4096];
    va_list va;

    va_start(va, fmt);
    vsnprintf(buf, sizeof(buf)-1, fmt, va);
    va_end(va);
    send(fd, buf, strlen(buf), 0);
}

//============================================================================

char *get_mime_type(char *name)
{
  char *ext = strrchr(name, '.');
  if (!ext) return NULL;
  if (strcmp(ext, ".html") == 0 || strcmp(ext, ".htm") == 0) return "text/html";
  if (strcmp(ext, ".jpg") == 0 || strcmp(ext, ".jpeg") == 0) return "image/jpeg";
  if (strcmp(ext, ".gif") == 0) return "image/gif";
  if (strcmp(ext, ".png") == 0) return "image/png";
  if (strcmp(ext, ".css") == 0) return "text/css";
  if (strcmp(ext, ".au") == 0) return "audio/basic";
  if (strcmp(ext, ".wav") == 0) return "audio/wav";
  if (strcmp(ext, ".avi") == 0) return "video/x-msvideo";
  if (strcmp(ext, ".mpeg") == 0 || strcmp(ext, ".mpg") == 0) return "video/mpeg";
  if (strcmp(ext, ".mp3") == 0) return "audio/mpeg";
  return NULL;
}

void send_headers(int fd, int status, char *title, char *extra, char *mime, 
                  int length, time_t date)
{
  time_t now;
  char timebuf[128];

  __fprintf(fd, "%s %d %s\r\n", PROTOCOL, status, title);
  __fprintf(fd, "Server: %s\r\n", SERVER);
  now = time(NULL);
  strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&now));
  __fprintf(fd, "Date: %s\r\n", timebuf);
  if (extra) __fprintf(fd, "%s\r\n", extra);
  if (mime) __fprintf(fd, "Content-Type: %s\r\n", mime);
  if (length >= 0) __fprintf(fd, "Content-Length: %d\r\n", length);
  if (date != -1)
  {
    strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&date));
    __fprintf(fd, "Last-Modified: %s\r\n", timebuf);
  }
  __fprintf(fd, "Connection: close\r\n");
  __fprintf(fd, "\r\n");
}

void send_error(int fd, int status, char *title, char *extra, char *text)
{
  send_headers(fd, status, title, extra, "text/html", -1, -1);
  __fprintf(fd, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\r\n", status, title);
  __fprintf(fd, "<BODY><H4>%d %s</H4>\r\n", status, title);
  __fprintf(fd, "%s\r\n", text);
  __fprintf(fd, "</BODY></HTML>\r\n");
}

void send_file(int fd, char *path, struct stat *statbuf)
{
  char data[4096];
  int n;

  FILE *file = fopen(path, "r");
  if (!file)
    send_error(fd, 403, "Forbidden", NULL, "Access denied.");
  else
  {
    int length = S_ISREG(statbuf->st_mode) ? statbuf->st_size : -1;
    send_headers(fd, 200, "OK", NULL, get_mime_type(path), length, statbuf->st_mtime);

    while ((n = fread(data, 1, sizeof(data), file)) > 0) if (write(fd,data,sizeof(data)) != sizeof(data)) perror("write");
    fclose(file);
  }
}

int process(int fd)
{
  char buf[4096];
  char *method;
  char *path;
  char *protocol;
  struct stat statbuf;
  char pathbuf[4096];
  int len;
  FILE *f;

  f = fdopen(fd, "a+");
  if (!fgets(buf, sizeof(buf), f)) return -1;
  printf("URL: %s", buf);

  method = strtok(buf, " ");
  path = strtok(NULL, " ");
  protocol = strtok(NULL, "\r");
  if (!method || !path || !protocol) return -1;

  fseek(f, 0, SEEK_CUR); // Force change of stream direction

  if (strcasecmp(method, "GET") != 0)
    send_error(fd, 501, "Not supported", NULL, "Method is not supported.");
  else if (stat(path, &statbuf) < 0)
    send_error(fd, 404, "Not Found", NULL, "File not found.");
  else if (S_ISDIR(statbuf.st_mode))
  {
    len = strlen(path);
    if (len == 0 || path[len - 1] != '/')
    {
      snprintf(pathbuf, sizeof(pathbuf), "Location: %s/", path);
      send_error(fd, 302, "Found", pathbuf, "Directories must end with a slash.");
    }
    else
    {
      snprintf(pathbuf, sizeof(pathbuf), "%sindex.html", path);
      if (stat(pathbuf, &statbuf) >= 0)
        send_file(fd, pathbuf, &statbuf);
      else
      {
        DIR *dir;
        struct dirent *de;

        send_headers(fd, 200, "OK", NULL, "text/html", -1, statbuf.st_mtime);
        __fprintf(fd, "<HTML><HEAD><TITLE>Index of %s</TITLE></HEAD>\r\n<BODY>", path);
        __fprintf(fd, "<H4>Index of %s</H4>\r\n<PRE>\n", path);
        __fprintf(fd, "Name                             Last Modified              Size\r\n");
        __fprintf(fd, "<HR>\r\n");
        if (len > 1) __fprintf(fd, "<A HREF=\"..\">..</A>\r\n");

        dir = opendir(path);
        while ((de = readdir(dir)) != NULL)
        {
          char timebuf[32];
          struct tm *tm;

          /* ignore dot(.) and dotdot(..) */
          if (strcmp( de->d_name, ".") == 0 || strcmp( de->d_name, ".." ) == 0) continue;

          strcpy(pathbuf, path);
          strcat(pathbuf, de->d_name);

          stat(pathbuf, &statbuf);
          tm = gmtime(&statbuf.st_mtime);
          strftime(timebuf, sizeof(timebuf), "%d-%b-%Y %H:%M:%S", tm);

          __fprintf(fd, "<A HREF=\"%s%s\">", de->d_name, S_ISDIR(statbuf.st_mode) ? "/" : "");
          __fprintf(fd, "%s%s", de->d_name, S_ISDIR(statbuf.st_mode) ? "/</A>" : "</A> ");
          if (strlen(de->d_name) < 32) __fprintf(fd, "%*s", 32 - strlen(de->d_name), "");

          if (S_ISDIR(statbuf.st_mode))
            __fprintf(fd, "%s\r\n", timebuf);
          else
            __fprintf(fd, "%s %10d\r\n", timebuf, statbuf.st_size);
        }
        closedir(dir);

        __fprintf(fd, "</PRE>\r\n<HR>\r\n<ADDRESS>%s</ADDRESS>\r\n</BODY></HTML>\r\n", SERVER);
      }
    }
  }
  else
    send_file(fd, path, &statbuf);

  fclose(f);
  return 0;
}

int main(void)
{
  int sock;
  struct sockaddr_in sin;

  sock = socket(AF_INET, SOCK_STREAM, 0);

  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = htons(PORT);
  bind(sock, (struct sockaddr *) &sin, sizeof(sin));

  listen(sock, 5);
  printf("HTTP server listening on port %d\n", PORT);

  while (1)
  {
    int s;

    s = accept(sock, NULL, NULL);
    if (s < 0) break;

    process(s);
  }

  close(sock);
  return 0;
}

